/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#ifndef CAOSDB_CONNECTION_H
#define CAOSDB_CONNECTION_H
/**
 * @file caosdb/connection.h
 * @author Timm Fitschen
 * @date 2021-05-18
 * @brief Configuration and setup of the connection.
 */
#include <iosfwd>                                // for ostream
#include <memory>                                // for shared_ptr, unique_ptr
#include <string>                                // for string
#include "caosdb/authentication.h"               // for Authenticator
#include "caosdb/entity/v1alpha1/main.grpc.pb.h" // for EntityTransactionSe...
#include "caosdb/info.h"                         // for VersionInfo
#include "caosdb/info/v1alpha1/main.grpc.pb.h"   // for GeneralInfoService:...
#include "caosdb/transaction.h"                  // for Transaction
#include "grpcpp/channel.h"                      // for Channel
#include "grpcpp/security/credentials.h"         // for ChannelCredentials

namespace caosdb::connection {
using caosdb::authentication::Authenticator;
using caosdb::entity::v1alpha1::EntityTransactionService;
using caosdb::info::VersionInfo;
using caosdb::info::v1alpha1::GeneralInfoService;
using caosdb::transaction::Transaction;
using grpc::ChannelCredentials;

class CertificateificateProvider {
public:
  [[nodiscard]] auto virtual GetCertificatePem() const -> std::string = 0;
};

class PemFileCertificateProvider : public CertificateificateProvider {
private:
  std::string cacert;

public:
  explicit PemFileCertificateProvider(const std::string &path);
  [[nodiscard]] auto GetCertificatePem() const -> std::string override;
};

class PemCertificateProvider : public CertificateificateProvider {
private:
  std::string cacert;

public:
  explicit PemCertificateProvider(const std::string &cacert);
  [[nodiscard]] auto GetCertificatePem() const -> std::string override;
};

/**
 * @brief Configuration of the CaosDB connection.
 */
class ConnectionConfig {
private:
  std::string host;
  int port;

public:
  ConnectionConfig(const std::string &host, int port);
  friend auto operator<<(std::ostream &out, const ConnectionConfig &config)
    -> std::ostream &;

  [[nodiscard]] auto virtual ToString() const -> std::string = 0;
  [[nodiscard]] auto GetHost() const -> std::string;
  [[nodiscard]] auto GetPort() const -> int;
  [[nodiscard]] auto virtual GetChannelCredentials() const
    -> std::shared_ptr<ChannelCredentials> = 0;
};

class InsecureConnectionConfig : public ConnectionConfig {
private:
  std::shared_ptr<ChannelCredentials> credentials;

public:
  InsecureConnectionConfig(const std::string &host, int port);
  [[nodiscard]] auto GetChannelCredentials() const
    -> std::shared_ptr<ChannelCredentials> override;
  [[nodiscard]] auto ToString() const -> std::string override;
};

class TlsConnectionConfig : public ConnectionConfig {
private:
  std::shared_ptr<ChannelCredentials> credentials;
  std::string cacert;

public:
  TlsConnectionConfig(const std::string &host, int port);
  TlsConnectionConfig(const std::string &host, int port,
                      const Authenticator &authenticator);
  TlsConnectionConfig(const std::string &host, int port,
                      const CertificateificateProvider &cacert);
  TlsConnectionConfig(const std::string &host, int port,
                      const CertificateificateProvider &cacert,
                      const Authenticator &authenticator);
  [[nodiscard]] auto GetChannelCredentials() const
    -> std::shared_ptr<ChannelCredentials> override;
  [[nodiscard]] auto ToString() const -> std::string override;
};

/**
 * @brief A reusable connection to a CaosDBServer.
 */
class Connection {
  std::shared_ptr<grpc::Channel> channel;
  std::unique_ptr<GeneralInfoService::Stub> general_info_service;
  std::shared_ptr<EntityTransactionService::Stub> entity_transaction_service;

public:
  explicit Connection(const ConnectionConfig &config);
  friend auto operator<<(std::ostream &out, const Connection &connection)
    -> std::ostream &;
  [[nodiscard]] auto GetVersionInfo() const -> std::unique_ptr<VersionInfo>;
  [[nodiscard]] auto CreateTransaction() const -> std::unique_ptr<Transaction>;
};
} // namespace caosdb::connection
#endif
